#include <sys/reboot.h>
#include <sys/types.h>
#include <sys/mount.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <limits.h>

int main(int argc, char **argv)
{
	FILE *f;
	char buf[4192];
	char err[4192];
	pid_t p, maxpid;
	int i;
	/* First of all, check which type of reboot/halt we want */
	int rb=RB_AUTOBOOT;
	if(!strcasecmp(argv[0], "poweroff"))
		rb=RB_POWER_OFF;
	else if(!strcasecmp(argv[0], "halt"))
		rb=RB_HALT_SYSTEM;

	printf("Killing all processes...");
	kill(-1, SIGTERM);
	sleep(2);
 
	printf("Killing all processes...");
	kill(-1, SIGKILL);
	sleep(2);
				
	printf("Syncing disks...\n");
	sync();

	printf("Remounting filesystems read-only...\n");

	f = fopen("/proc/mounts", "r");

	while(fgets(buf, 4192, f)) {
		char *fs=strtok(buf, " \t");
		char *mp=strtok(NULL, " \t");
		char *fstype=strtok(NULL, " \t");
		sprintf(err, "Remounting %s (%s) (%s)", fs, mp, fstype);
		if(mount(fs, mp, fstype, MS_MGC_VAL|MS_REMOUNT|MS_SYNCHRONOUS|MS_RDONLY, NULL))
			perror(err);
	}
	fclose(f);
	
	pivot_root("/initrd", "mnt/source");
	chdir("/");
	
	printf("Unmounting filesystems...\n");
	for(i=0; i<10; i++) {
		/* We do this several times because we can't unmount
		 * /usr while /usr/local is mounted (fs busy); the first pass
		 * will unmount /usr/local, the second will catch /usr... */
		f=fopen("/proc/mounts", "r");
		while(fgets(buf, 4192, f)) {
			char *fs=strtok(buf, " \t");
			char *mp=strtok(NULL, " \t");
			char *fstype=strtok(NULL, " \t");
			if(!strcmp(fs, "/") || !strcmp(fs, "/proc"))
				continue;
			sprintf(err, "Unmounting %s (%s) (%s)", fs, mp, fstype);
			if(umount(mp))
				perror(err);
		}
		fclose(f);
	}
	if(umount("/proc"))
		perror("Unmounting /proc");
	if(umount("/"))
		perror("Unmounting /");
	
	reboot(rb);
}
